@@ -40,12 +40,14 @@ import Control.Monad
          ( filterM, liftM )
 import qualified System.Directory as System
          ( doesFileExist, doesDirectoryExist )
+import qualified Data.Map as Map
 
 import Distribution.PackageDescription
 import Distribution.PackageDescription.Configuration
          ( flattenPackageDescription, finalizePackageDescription )
 import Distribution.Compiler
-         ( CompilerFlavor(..), buildCompilerFlavor, CompilerId(..) )
+         ( CompilerFlavor(..), buildCompilerFlavor, CompilerId(..)
+         , unknownCompilerInfo, AbiTag(..) )
 import Distribution.System
          ( OS(..), Arch(..), buildPlatform )
 import Distribution.License
@@ -73,7 +75,8 @@ import Text.PrettyPrint ((<>), (<+>))
 
 import qualified Language.Haskell.Extension as Extension (deprecatedExtensions)
 import Language.Haskell.Extension
-         ( Language(UnknownLanguage), knownLanguages, Extension(..), KnownExtension(..) )
+         ( Language(UnknownLanguage), knownLanguages
+         , Extension(..), KnownExtension(..) )
 import System.FilePath
          ( (</>), takeExtension, isRelative, isAbsolute
          , splitDirectories,  splitPath )
@@ -119,7 +122,8 @@ check :: Bool -> PackageCheck -> Maybe PackageCheck
 check False _  = Nothing
 check True  pc = Just pc
 
-checkSpecVersion :: PackageDescription -> [Int] -> Bool -> PackageCheck -> Maybe PackageCheck
+checkSpecVersion :: PackageDescription -> [Int] -> Bool -> PackageCheck
+                 -> Maybe PackageCheck
 checkSpecVersion pkg specver cond pc
   | specVersion pkg >= Version specver [] = Nothing
   | otherwise                             = check cond pc
@@ -189,7 +193,8 @@ checkSanity pkg =
         ++ ". The name of every executable, test suite, and benchmark section in"
         ++ " the package must be unique."
   ]
-  --TODO: check for name clashes case insensitively: windows file systems cannot cope.
+  --TODO: check for name clashes case insensitively: windows file systems cannot
+  --cope.
 
   ++ maybe []  (checkLibrary    pkg) (library pkg)
   ++ concatMap (checkExecutable pkg) (executables pkg)
@@ -211,17 +216,34 @@ checkSanity pkg =
     duplicateNames = dups $ exeNames ++ testNames ++ bmNames
 
 checkLibrary :: PackageDescription -> Library -> [PackageCheck]
-checkLibrary _pkg lib =
+checkLibrary pkg lib =
   catMaybes [
 
     check (not (null moduleDuplicates)) $
        PackageBuildImpossible $
             "Duplicate modules in library: "
          ++ commaSep (map display moduleDuplicates)
+
+    -- check use of required-signatures/exposed-signatures sections
+  , checkVersion [1,21] (not (null (requiredSignatures lib))) $
+      PackageDistInexcusable $
+           "To use the 'required-signatures' field the package needs to specify "
+        ++ "at least 'cabal-version: >= 1.21'."
+
+  , checkVersion [1,21] (not (null (exposedSignatures lib))) $
+      PackageDistInexcusable $
+           "To use the 'exposed-signatures' field the package needs to specify "
+        ++ "at least 'cabal-version: >= 1.21'."
   ]
 
   where
-    moduleDuplicates = dups (libModules lib)
+    checkVersion :: [Int] -> Bool -> PackageCheck -> Maybe PackageCheck
+    checkVersion ver cond pc
+      | specVersion pkg >= Version ver []      = Nothing
+      | otherwise                              = check cond pc
+
+    moduleDuplicates = dups (libModules lib ++
+                             map moduleReexportName (reexportedModules lib))
 
 checkExecutable :: PackageDescription -> Executable -> [PackageCheck]
 checkExecutable pkg exe =
@@ -473,10 +495,13 @@ checkLicense :: PackageDescription -> [PackageCheck]
 checkLicense pkg =
   catMaybes [
 
-    check (license pkg == AllRightsReserved) $
+    check (license pkg == UnspecifiedLicense) $
       PackageDistInexcusable
-        "The 'license' field is missing or specified as AllRightsReserved."
+        "The 'license' field is missing."
 
+  , check (license pkg == AllRightsReserved) $
+      PackageDistSuspicious
+        "The 'license' is AllRightsReserved. Is that really what you want?"
   , case license pkg of
       UnknownLicense l -> Just $
         PackageBuildWarning $
@@ -501,7 +526,8 @@ checkLicense pkg =
           ++ "version then please file a ticket."
       _ -> Nothing
 
-  , check (license pkg `notElem` [AllRightsReserved, PublicDomain]
+  , check (license pkg `notElem` [ AllRightsReserved
+                                 , UnspecifiedLicense, PublicDomain]
            -- AllRightsReserved and PublicDomain are not strictly
            -- licenses so don't need license files.
         && null (licenseFiles pkg)) $
@@ -595,9 +621,12 @@ checkGhcOptions pkg =
       PackageDistInexcusable $
         "'ghc-options: -fhpc' is not appropriate for a distributed package."
 
-  , check (any ("-d" `isPrefixOf`) all_ghc_options) $
+    -- -dynamic is not a debug flag
+  , check (any (\opt -> "-d" `isPrefixOf` opt && opt /= "-dynamic")
+           all_ghc_options) $
       PackageDistInexcusable $
-        "'ghc-options: -d*' debug flags are not appropriate for a distributed package."
+        "'ghc-options: -d*' debug flags are not appropriate "
+        ++ "for a distributed package."
 
   , checkFlags ["-prof"] $
       PackageBuildWarning $
@@ -607,37 +636,43 @@ checkGhcOptions pkg =
 
   , checkFlags ["-o"] $
       PackageBuildWarning $
-        "'ghc-options: -o' is not needed. The output files are named automatically."
+           "'ghc-options: -o' is not needed. "
+        ++ "The output files are named automatically."
 
   , checkFlags ["-hide-package"] $
       PackageBuildWarning $
-           "'ghc-options: -hide-package' is never needed. Cabal hides all packages."
+      "'ghc-options: -hide-package' is never needed. "
+      ++ "Cabal hides all packages."
 
   , checkFlags ["--make"] $
       PackageBuildWarning $
-        "'ghc-options: --make' is never needed. Cabal uses this automatically."
+      "'ghc-options: --make' is never needed. Cabal uses this automatically."
 
   , checkFlags ["-main-is"] $
       PackageDistSuspicious $
-           "'ghc-options: -main-is' is not portable."
+      "'ghc-options: -main-is' is not portable."
 
   , checkFlags ["-O0", "-Onot"] $
       PackageDistSuspicious $
-        "'ghc-options: -O0' is not needed. Use the --disable-optimization configure flag."
+      "'ghc-options: -O0' is not needed. "
+      ++ "Use the --disable-optimization configure flag."
 
   , checkFlags [ "-O", "-O1"] $
       PackageDistInexcusable $
-           "'ghc-options: -O' is not needed. Cabal automatically adds the '-O' flag. "
-        ++ "Setting it yourself interferes with the --disable-optimization flag."
+      "'ghc-options: -O' is not needed. "
+      ++ "Cabal automatically adds the '-O' flag. "
+      ++ "Setting it yourself interferes with the --disable-optimization flag."
 
   , checkFlags ["-O2"] $
       PackageDistSuspicious $
-           "'ghc-options: -O2' is rarely needed. Check that it is giving a real benefit "
-        ++ "and not just imposing longer compile times on your users."
+      "'ghc-options: -O2' is rarely needed. "
+      ++ "Check that it is giving a real benefit "
+      ++ "and not just imposing longer compile times on your users."
 
   , checkFlags ["-split-objs"] $
       PackageBuildWarning $
-        "'ghc-options: -split-objs' is not needed. Use the --enable-split-objs configure flag."
+        "'ghc-options: -split-objs' is not needed. "
+        ++ "Use the --enable-split-objs configure flag."
 
   , checkFlags ["-optl-Wl,-s", "-optl-s"] $
       PackageDistInexcusable $
@@ -649,7 +684,13 @@ checkGhcOptions pkg =
 
   , checkFlags ["-fglasgow-exts"] $
       PackageDistSuspicious $
-        "Instead of 'ghc-options: -fglasgow-exts' it is preferable to use the 'extensions' field."
+        "Instead of 'ghc-options: -fglasgow-exts' it is preferable to use "
+        ++ "the 'extensions' field."
+
+  , checkProfFlags ["-auto-all"] $
+      PackageDistSuspicious $
+        "'ghc-prof-options: -auto-all' is fine during development, but "
+        ++ "not recommended in a distributed package. "
 
   , check ("-threaded" `elem` lib_ghc_options) $
       PackageDistSuspicious $
@@ -683,49 +724,57 @@ checkGhcOptions pkg =
                            && ("-Wall"   `elem` opts || "-W" `elem` opts)
     has_Werror     = any (\opts -> "-Werror" `elem` opts) ghc_options
 
-    ghc_options = [ strs | bi <- allBuildInfo pkg
-                         , (GHC, strs) <- options bi ]
-    all_ghc_options = concat ghc_options
+    (ghc_options, ghc_prof_options) =
+      unzip . map (\bi -> (hcOptions GHC bi, hcProfOptions GHC bi))
+      $ (allBuildInfo pkg)
+    all_ghc_options      = concat ghc_options
+    all_ghc_prof_options = concat ghc_prof_options
     lib_ghc_options = maybe [] (hcOptions GHC . libBuildInfo) (library pkg)
 
-    checkFlags :: [String] -> PackageCheck -> Maybe PackageCheck
-    checkFlags flags = check (any (`elem` flags) all_ghc_options)
+    checkFlags,checkProfFlags :: [String] -> PackageCheck -> Maybe PackageCheck
+    checkFlags     flags = doCheckFlags flags all_ghc_options
+    checkProfFlags flags = doCheckFlags flags all_ghc_prof_options
+
+    doCheckFlags   flags opts = check (any (`elem` flags) opts)
 
     ghcExtension ('-':'f':name) = case name of
-      "allow-overlapping-instances"    -> Just (EnableExtension  OverlappingInstances)
-      "no-allow-overlapping-instances" -> Just (DisableExtension OverlappingInstances)
-      "th"                             -> Just (EnableExtension  TemplateHaskell)
-      "no-th"                          -> Just (DisableExtension TemplateHaskell)
-      "ffi"                            -> Just (EnableExtension  ForeignFunctionInterface)
-      "no-ffi"                         -> Just (DisableExtension ForeignFunctionInterface)
-      "fi"                             -> Just (EnableExtension  ForeignFunctionInterface)
-      "no-fi"                          -> Just (DisableExtension ForeignFunctionInterface)
-      "monomorphism-restriction"       -> Just (EnableExtension  MonomorphismRestriction)
-      "no-monomorphism-restriction"    -> Just (DisableExtension MonomorphismRestriction)
-      "mono-pat-binds"                 -> Just (EnableExtension  MonoPatBinds)
-      "no-mono-pat-binds"              -> Just (DisableExtension MonoPatBinds)
-      "allow-undecidable-instances"    -> Just (EnableExtension  UndecidableInstances)
-      "no-allow-undecidable-instances" -> Just (DisableExtension UndecidableInstances)
-      "allow-incoherent-instances"     -> Just (EnableExtension  IncoherentInstances)
-      "no-allow-incoherent-instances"  -> Just (DisableExtension IncoherentInstances)
-      "arrows"                         -> Just (EnableExtension  Arrows)
-      "no-arrows"                      -> Just (DisableExtension Arrows)
-      "generics"                       -> Just (EnableExtension  Generics)
-      "no-generics"                    -> Just (DisableExtension Generics)
-      "implicit-prelude"               -> Just (EnableExtension  ImplicitPrelude)
-      "no-implicit-prelude"            -> Just (DisableExtension ImplicitPrelude)
-      "implicit-params"                -> Just (EnableExtension  ImplicitParams)
-      "no-implicit-params"             -> Just (DisableExtension ImplicitParams)
-      "bang-patterns"                  -> Just (EnableExtension  BangPatterns)
-      "no-bang-patterns"               -> Just (DisableExtension BangPatterns)
-      "scoped-type-variables"          -> Just (EnableExtension  ScopedTypeVariables)
-      "no-scoped-type-variables"       -> Just (DisableExtension ScopedTypeVariables)
-      "extended-default-rules"         -> Just (EnableExtension  ExtendedDefaultRules)
-      "no-extended-default-rules"      -> Just (DisableExtension ExtendedDefaultRules)
+      "allow-overlapping-instances"    -> enable  OverlappingInstances
+      "no-allow-overlapping-instances" -> disable OverlappingInstances
+      "th"                             -> enable  TemplateHaskell
+      "no-th"                          -> disable TemplateHaskell
+      "ffi"                            -> enable  ForeignFunctionInterface
+      "no-ffi"                         -> disable ForeignFunctionInterface
+      "fi"                             -> enable  ForeignFunctionInterface
+      "no-fi"                          -> disable ForeignFunctionInterface
+      "monomorphism-restriction"       -> enable  MonomorphismRestriction
+      "no-monomorphism-restriction"    -> disable MonomorphismRestriction
+      "mono-pat-binds"                 -> enable  MonoPatBinds
+      "no-mono-pat-binds"              -> disable MonoPatBinds
+      "allow-undecidable-instances"    -> enable  UndecidableInstances
+      "no-allow-undecidable-instances" -> disable UndecidableInstances
+      "allow-incoherent-instances"     -> enable  IncoherentInstances
+      "no-allow-incoherent-instances"  -> disable IncoherentInstances
+      "arrows"                         -> enable  Arrows
+      "no-arrows"                      -> disable Arrows
+      "generics"                       -> enable  Generics
+      "no-generics"                    -> disable Generics
+      "implicit-prelude"               -> enable  ImplicitPrelude
+      "no-implicit-prelude"            -> disable ImplicitPrelude
+      "implicit-params"                -> enable  ImplicitParams
+      "no-implicit-params"             -> disable ImplicitParams
+      "bang-patterns"                  -> enable  BangPatterns
+      "no-bang-patterns"               -> disable BangPatterns
+      "scoped-type-variables"          -> enable  ScopedTypeVariables
+      "no-scoped-type-variables"       -> disable ScopedTypeVariables
+      "extended-default-rules"         -> enable  ExtendedDefaultRules
+      "no-extended-default-rules"      -> disable ExtendedDefaultRules
       _                                -> Nothing
-    ghcExtension "-cpp"             = Just (EnableExtension CPP)
+    ghcExtension "-cpp"             = enable CPP
     ghcExtension _                  = Nothing
 
+    enable  e = Just (EnableExtension e)
+    disable e = Just (DisableExtension e)
+
 checkCCOptions :: PackageDescription -> [PackageCheck]
 checkCCOptions pkg =
   catMaybes [
@@ -830,6 +879,7 @@ checkPaths pkg =
       ++ [ (path, "data-dir")        | path <- [dataDir      pkg]]
       ++ concat
          [    [ (path, "c-sources")        | path <- cSources        bi ]
+           ++ [ (path, "js-sources")       | path <- jsSources       bi ]
            ++ [ (path, "install-includes") | path <- installIncludes bi ]
            ++ [ (path, "hs-source-dirs")   | path <- hsSourceDirs    bi ]
          | bi <- allBuildInfo pkg ]
@@ -902,6 +952,22 @@ checkCabalVersion pkg =
         ++ "different modules then list the other ones in the "
         ++ "'other-languages' field."
 
+    -- check use of reexported-modules sections
+  , checkVersion [1,21]
+    (maybe False (not.null.reexportedModules) (library pkg)) $
+      PackageDistInexcusable $
+           "To use the 'reexported-module' field the package needs to specify "
+        ++ "at least 'cabal-version: >= 1.21'."
+
+    -- check use of thinning and renaming
+  , checkVersion [1,21] (not (null depsUsingThinningRenamingSyntax)) $
+      PackageDistInexcusable $
+           "The package uses "
+        ++ "thinning and renaming in the 'build-depends' field: "
+        ++ commaSep (map display depsUsingThinningRenamingSyntax)
+        ++ ". To use this new syntax, the package needs to specify at least"
+        ++ "'cabal-version: >= 1.21'."
+
     -- check use of default-extensions field
     -- don't need to do the equivalent check for other-extensions
   , checkVersion [1,10] (any (not . null) (buildInfoField defaultExtensions)) $
@@ -1072,9 +1138,19 @@ checkCabalVersion pkg =
     depsUsingWildcardSyntax = [ dep | dep@(Dependency _ vr) <- buildDepends pkg
                                     , usesWildcardSyntax vr ]
 
-    testedWithUsingWildcardSyntax = [ Dependency (PackageName (display compiler)) vr
-                                    | (compiler, vr) <- testedWith pkg
-                                    , usesWildcardSyntax vr ]
+    -- XXX: If the user writes build-depends: foo with (), this is
+    -- indistinguishable from build-depends: foo, so there won't be an
+    -- error even though there should be
+    depsUsingThinningRenamingSyntax =
+      [ name
+      | bi <- allBuildInfo pkg
+      , (name, rns) <- Map.toList (targetBuildRenaming bi)
+      , rns /= ModuleRenaming True [] ]
+
+    testedWithUsingWildcardSyntax =
+      [ Dependency (PackageName (display compiler)) vr
+      | (compiler, vr) <- testedWith pkg
+      , usesWildcardSyntax vr ]
 
     usesWildcardSyntax :: VersionRange -> Bool
     usesWildcardSyntax =
@@ -1094,7 +1170,8 @@ checkCabalVersion pkg =
         intersectVersionRanges unionVersionRanges id
 
     compatLicenses = [ GPL Nothing, LGPL Nothing, AGPL Nothing, BSD3, BSD4
-                     , PublicDomain, AllRightsReserved, OtherLicense ]
+                     , PublicDomain, AllRightsReserved
+                     , UnspecifiedLicense, OtherLicense ]
 
     mentionedExtensions = [ ext | bi <- allBuildInfo pkg
                                 , ext <- allExtensions bi ]
@@ -1154,8 +1231,10 @@ displayRawVersionRange =
      (\v   -> (Disp.text ">=" <> disp v                   , 0))
      (\v   -> (Disp.text "<=" <> disp v                   , 0))
      (\v _ -> (Disp.text "==" <> dispWild v               , 0))
-     (\(r1, p1) (r2, p2) -> (punct 2 p1 r1 <+> Disp.text "||" <+> punct 2 p2 r2 , 2))
-     (\(r1, p1) (r2, p2) -> (punct 1 p1 r1 <+> Disp.text "&&" <+> punct 1 p2 r2 , 1))
+     (\(r1, p1) (r2, p2) ->
+       (punct 2 p1 r1 <+> Disp.text "||" <+> punct 2 p2 r2 , 2))
+     (\(r1, p1) (r2, p2) ->
+       (punct 1 p1 r1 <+> Disp.text "&&" <+> punct 1 p2 r2 , 1))
      (\(r,  _ )          -> (Disp.parens r, 0)) -- parens
 
   where
@@ -1208,7 +1287,8 @@ checkPackageVersions pkg =
     -- using no package index and the current platform.
     finalised = finalizePackageDescription
                               [] (const True) buildPlatform
-                              (CompilerId buildCompilerFlavor (Version [] []))
+                              (unknownCompilerInfo
+                                (CompilerId buildCompilerFlavor (Version [] [])) NoAbiTag)
                               [] pkg
     baseDependency = case finalised of
       Right (pkg', _) | not (null baseDeps) ->
@@ -1341,7 +1421,8 @@ checkConfigureExists ops PackageDescription { buildType = Just Configure } = do
   exists <- doesFileExist ops "configure"
   return $ check (not exists) $
     PackageBuildWarning $
-      "The 'build-type' is 'Configure' but there is no 'configure' script."
+      "The 'build-type' is 'Configure' but there is no 'configure' script. "
+      ++ "You probably need to run 'autoreconf -i' to generate it."
 checkConfigureExists _ _ = return Nothing
 
 checkLocalPathsExist :: Monad m => CheckPackageContentOps m
