@@ -50,7 +50,8 @@ import Distribution.ReadE
 import Distribution.Text
          ( Text(..) )
 import Distribution.Simple.Utils
-         ( comparing, intercalate, lowercase, normaliseLineEndings )
+         ( comparing, dropWhileEndLE, intercalate, lowercase
+         , normaliseLineEndings )
 import Language.Haskell.Extension
          ( Language, Extension )
 
@@ -198,7 +199,7 @@ commaListFieldWithSep separator name showF readF get set =
    where
      set' xs b = set (get b ++ xs) b
      showF'    = separator . punctuate comma . map showF
- 
+
 commaListField :: String -> (a -> Doc) -> ReadP [a] a
                  -> (b -> [a]) -> ([a] -> b -> b) -> FieldDescr b
 commaListField = commaListFieldWithSep fsep
@@ -267,7 +268,7 @@ ppFields fields x =
    vcat [ ppField name (getter x) | FieldDescr name getter _ <- fields ]
 
 ppField :: String -> Doc -> Doc
-ppField name fielddoc 
+ppField name fielddoc
    | isEmpty fielddoc         = empty
    | name `elem` nestedFields = text name <> colon $+$ nest indentWith fielddoc
    | otherwise                = text name <> colon <+> fielddoc
@@ -280,6 +281,7 @@ ppField name fielddoc
          , "extra-tmp-files"
          , "exposed-modules"
          , "c-sources"
+         , "js-sources"
          , "extra-libraries"
          , "includes"
          , "install-includes"
@@ -477,7 +479,7 @@ tokeniseLineFlat (n0, i, t, l)
 
 trimLeading, trimTrailing :: String -> String
 trimLeading  = dropWhile isSpace
-trimTrailing = reverse . dropWhile isSpace . reverse
+trimTrailing = dropWhileEndLE isSpace
 
 
 type SyntaxTree = Tree (LineNo, HasTabs, String)
@@ -656,7 +658,7 @@ parseOptVersion :: ReadP r Version
 parseOptVersion = parseQuoted ver <++ ver
   where ver :: ReadP r Version
         ver = parse <++ return noVersion
-        noVersion = Version{ versionBranch=[], versionTags=[] }
+        noVersion = Version [] []
 
 parseTestedWithQ :: ReadP r (CompilerFlavor,VersionRange)
 parseTestedWithQ = parseQuoted tw <++ tw
