@@ -1,4 +1,6 @@
 {-# LANGUAGE CPP, ForeignFunctionInterface #-}
+{-# LANGUAGE DeriveGeneric #-}
+
 -----------------------------------------------------------------------------
 -- |
 -- Module      :  Distribution.Simple.InstallDirs
@@ -37,23 +39,26 @@ module Distribution.Simple.InstallDirs (
         platformTemplateEnv,
         compilerTemplateEnv,
         packageTemplateEnv,
+        abiTemplateEnv,
         installDirsTemplateEnv,
   ) where
 
 
+import Distribution.Compat.Binary (Binary)
 import Data.List (isPrefixOf)
 import Data.Maybe (fromMaybe)
 import Data.Monoid (Monoid(..))
+import GHC.Generics (Generic)
 import System.Directory (getAppUserDataDirectory)
 import System.FilePath ((</>), isPathSeparator, pathSeparator)
 import System.FilePath (dropDrive)
 
 import Distribution.Package
-         ( PackageIdentifier, packageName, packageVersion )
+         ( PackageIdentifier, PackageKey, packageName, packageVersion )
 import Distribution.System
          ( OS(..), buildOS, Platform(..) )
 import Distribution.Compiler
-         ( CompilerId, CompilerFlavor(..) )
+         ( AbiTag(..), abiTagString, CompilerInfo(..), CompilerFlavor(..) )
 import Distribution.Text
          ( display )
 
@@ -80,7 +85,6 @@ data InstallDirs dir = InstallDirs {
         libsubdir    :: dir,
         dynlibdir    :: dir,
         libexecdir   :: dir,
-        progdir      :: dir,
         includedir   :: dir,
         datadir      :: dir,
         datasubdir   :: dir,
@@ -89,7 +93,9 @@ data InstallDirs dir = InstallDirs {
         htmldir      :: dir,
         haddockdir   :: dir,
         sysconfdir   :: dir
-    } deriving (Read, Show)
+    } deriving (Generic, Read, Show)
+
+instance Binary dir => Binary (InstallDirs dir)
 
 instance Functor InstallDirs where
   fmap f dirs = InstallDirs {
@@ -99,7 +105,6 @@ instance Functor InstallDirs where
     libsubdir    = f (libsubdir dirs),
     dynlibdir    = f (dynlibdir dirs),
     libexecdir   = f (libexecdir dirs),
-    progdir      = f (progdir dirs),
     includedir   = f (includedir dirs),
     datadir      = f (datadir dirs),
     datasubdir   = f (datasubdir dirs),
@@ -118,7 +123,6 @@ instance Monoid dir => Monoid (InstallDirs dir) where
       libsubdir    = mempty,
       dynlibdir    = mempty,
       libexecdir   = mempty,
-      progdir      = mempty,
       includedir   = mempty,
       datadir      = mempty,
       datasubdir   = mempty,
@@ -141,7 +145,6 @@ combineInstallDirs combine a b = InstallDirs {
     libsubdir    = libsubdir a  `combine` libsubdir b,
     dynlibdir    = dynlibdir a  `combine` dynlibdir b,
     libexecdir   = libexecdir a `combine` libexecdir b,
-    progdir      = progdir a    `combine` progdir b,
     includedir   = includedir a `combine` includedir b,
     datadir      = datadir a    `combine` datadir b,
     datasubdir   = datasubdir a `combine` datasubdir b,
@@ -177,7 +180,7 @@ appendSubdirs append dirs = dirs {
 -- users to be able to configure @--libdir=\/usr\/lib64@ for example but
 -- because by default we want to support installing multiple versions of
 -- packages and building the same package for multiple compilers we append the
--- libsubdir to get: @\/usr\/lib64\/$pkgid\/$compiler@.
+-- libsubdir to get: @\/usr\/lib64\/$pkgkey\/$compiler@.
 --
 -- An additional complication is the need to support relocatable packages on
 -- systems which support such things, like Windows.
@@ -207,22 +210,20 @@ defaultInstallDirs comp userInstall _hasLibs = do
       bindir       = "$prefix" </> "bin",
       libdir       = installLibDir,
       libsubdir    = case comp of
-           Hugs   -> "hugs" </> "packages" </> "$pkg"
            JHC    -> "$compiler"
            LHC    -> "$compiler"
            UHC    -> "$pkgid"
-           _other -> "$arch-$os-$compiler" </> "$pkgid",
+           _other -> "$abi" </> "$pkgkey",
       dynlibdir    = "$libdir",
       libexecdir   = case buildOS of
-        Windows   -> "$prefix" </> "$pkgid"
+        Windows   -> "$prefix" </> "$pkgkey"
         _other    -> "$prefix" </> "libexec",
-      progdir      = "$libdir" </> "hugs" </> "programs",
       includedir   = "$libdir" </> "$libsubdir" </> "include",
       datadir      = case buildOS of
         Windows   -> "$prefix"
         _other    -> "$prefix" </> "share",
-      datasubdir   = "$arch-$os-$compiler" </> "$pkgid",
-      docdir       = "$datadir" </> "doc" </> "$arch-$os-$compiler" </> "$pkgid",
+      datasubdir   = "$abi" </> "$pkgid",
+      docdir       = "$datadir" </> "doc" </> "$abi" </> "$pkgid",
       mandir       = "$datadir" </> "man",
       htmldir      = "$docdir"  </> "html",
       haddockdir   = "$htmldir",
@@ -256,7 +257,6 @@ substituteInstallDirTemplates env dirs = dirs'
       libsubdir  = subst libsubdir  [],
       dynlibdir  = subst dynlibdir  [prefixVar, bindirVar, libdirVar],
       libexecdir = subst libexecdir prefixBinLibVars,
-      progdir    = subst progdir    prefixBinLibVars,
       includedir = subst includedir prefixBinLibVars,
       datadir    = subst datadir    prefixBinLibVars,
       datasubdir = subst datasubdir [],
@@ -283,10 +283,14 @@ substituteInstallDirTemplates env dirs = dirs'
 -- | Convert from abstract install directories to actual absolute ones by
 -- substituting for all the variables in the abstract paths, to get real
 -- absolute path.
-absoluteInstallDirs :: PackageIdentifier -> CompilerId -> CopyDest -> Platform
+absoluteInstallDirs :: PackageIdentifier
+                    -> PackageKey
+                    -> CompilerInfo
+                    -> CopyDest
+                    -> Platform
                     -> InstallDirs PathTemplate
                     -> InstallDirs FilePath
-absoluteInstallDirs pkgId compilerId copydest platform dirs =
+absoluteInstallDirs pkgId pkg_key compilerId copydest platform dirs =
     (case copydest of
        CopyTo destdir -> fmap ((destdir </>) . dropDrive)
        _              -> id)
@@ -294,7 +298,7 @@ absoluteInstallDirs pkgId compilerId copydest platform dirs =
   . fmap fromPathTemplate
   $ substituteInstallDirTemplates env dirs
   where
-    env = initialPathTemplateEnv pkgId compilerId platform
+    env = initialPathTemplateEnv pkgId pkg_key compilerId platform
 
 
 -- |The location prefix for the /copy/ command.
@@ -309,10 +313,13 @@ data CopyDest
 -- prevents us from making a relocatable package (also known as a \"prefix
 -- independent\" package).
 --
-prefixRelativeInstallDirs :: PackageIdentifier -> CompilerId -> Platform
+prefixRelativeInstallDirs :: PackageIdentifier
+                          -> PackageKey
+                          -> CompilerInfo
+                          -> Platform
                           -> InstallDirTemplates
                           -> InstallDirs (Maybe FilePath)
-prefixRelativeInstallDirs pkgId compilerId platform dirs =
+prefixRelativeInstallDirs pkgId pkg_key compilerId platform dirs =
     fmap relative
   . appendSubdirs combinePathTemplate
   $ -- substitute the path template into each other, except that we map
@@ -322,7 +329,7 @@ prefixRelativeInstallDirs pkgId compilerId platform dirs =
       prefix = PathTemplate [Variable PrefixVar]
     }
   where
-    env = initialPathTemplateEnv pkgId compilerId platform
+    env = initialPathTemplateEnv pkgId pkg_key compilerId platform
 
     -- If it starts with $prefix then it's relative and produce the relative
     -- path by stripping off $prefix/ or $prefix
@@ -339,12 +346,16 @@ prefixRelativeInstallDirs pkgId compilerId platform dirs =
 -- | An abstract path, possibly containing variables that need to be
 -- substituted for to get a real 'FilePath'.
 --
-newtype PathTemplate = PathTemplate [PathComponent]
+newtype PathTemplate = PathTemplate [PathComponent] deriving (Eq, Generic, Ord)
+
+instance Binary PathTemplate
 
 data PathComponent =
        Ordinary FilePath
      | Variable PathTemplateVariable
-     deriving Eq
+     deriving (Eq, Ord, Generic)
+
+instance Binary PathComponent
 
 data PathTemplateVariable =
        PrefixVar     -- ^ The @$prefix@ path variable
@@ -358,15 +369,20 @@ data PathTemplateVariable =
      | PkgNameVar    -- ^ The @$pkg@ package name path variable
      | PkgVerVar     -- ^ The @$version@ package version path variable
      | PkgIdVar      -- ^ The @$pkgid@ package Id path variable, eg @foo-1.0@
+     | PkgKeyVar     -- ^ The @$pkgkey@ package key path variable
      | CompilerVar   -- ^ The compiler name and version, eg @ghc-6.6.1@
      | OSVar         -- ^ The operating system name, eg @windows@ or @linux@
      | ArchVar       -- ^ The CPU architecture name, eg @i386@ or @x86_64@
+     | AbiVar        -- ^ The Compiler's ABI identifier, $arch-$os-$compiler-$abitag
+     | AbiTagVar     -- ^ The optional ABI tag for the compiler
      | ExecutableNameVar -- ^ The executable name; used in shell wrappers
      | TestSuiteNameVar   -- ^ The name of the test suite being run
      | TestSuiteResultVar -- ^ The result of the test suite being run, eg
                           -- @pass@, @fail@, or @error@.
      | BenchmarkNameVar   -- ^ The name of the benchmark being run
-  deriving Eq
+  deriving (Eq, Ord, Generic)
+
+instance Binary PathTemplateVariable
 
 type PathTemplateEnv = [(PathTemplateVariable, PathTemplate)]
 
@@ -395,23 +411,28 @@ substPathTemplate environment (PathTemplate template) =
                   Nothing                        -> [component]
 
 -- | The initial environment has all the static stuff but no paths
-initialPathTemplateEnv :: PackageIdentifier -> CompilerId -> Platform
+initialPathTemplateEnv :: PackageIdentifier
+                       -> PackageKey
+                       -> CompilerInfo
+                       -> Platform
                        -> PathTemplateEnv
-initialPathTemplateEnv pkgId compilerId platform =
-     packageTemplateEnv  pkgId
-  ++ compilerTemplateEnv compilerId
+initialPathTemplateEnv pkgId pkg_key compiler platform =
+     packageTemplateEnv  pkgId pkg_key
+  ++ compilerTemplateEnv compiler
   ++ platformTemplateEnv platform
+  ++ abiTemplateEnv compiler platform
 
-packageTemplateEnv :: PackageIdentifier -> PathTemplateEnv
-packageTemplateEnv pkgId =
+packageTemplateEnv :: PackageIdentifier -> PackageKey -> PathTemplateEnv
+packageTemplateEnv pkgId pkg_key =
   [(PkgNameVar,  PathTemplate [Ordinary $ display (packageName pkgId)])
   ,(PkgVerVar,   PathTemplate [Ordinary $ display (packageVersion pkgId)])
+  ,(PkgKeyVar,   PathTemplate [Ordinary $ display pkg_key])
   ,(PkgIdVar,    PathTemplate [Ordinary $ display pkgId])
   ]
 
-compilerTemplateEnv :: CompilerId -> PathTemplateEnv
-compilerTemplateEnv compilerId =
-  [(CompilerVar, PathTemplate [Ordinary $ display compilerId])
+compilerTemplateEnv :: CompilerInfo -> PathTemplateEnv
+compilerTemplateEnv compiler =
+  [(CompilerVar, PathTemplate [Ordinary $ display (compilerInfoId compiler)])
   ]
 
 platformTemplateEnv :: Platform -> PathTemplateEnv
@@ -420,6 +441,16 @@ platformTemplateEnv (Platform arch os) =
   ,(ArchVar,     PathTemplate [Ordinary $ display arch])
   ]
 
+abiTemplateEnv :: CompilerInfo -> Platform -> PathTemplateEnv
+abiTemplateEnv compiler (Platform arch os) =
+  [(AbiVar,      PathTemplate [Ordinary $ display arch ++ '-':display os ++
+                                          '-':display (compilerInfoId compiler) ++
+                                          case compilerInfoAbiTag compiler of
+                                            NoAbiTag   -> ""
+                                            AbiTag tag -> '-':tag])
+  ,(AbiTagVar,   PathTemplate [Ordinary $ abiTagString (compilerInfoAbiTag compiler)])
+  ]
+
 installDirsTemplateEnv :: InstallDirs PathTemplate -> PathTemplateEnv
 installDirsTemplateEnv dirs =
   [(PrefixVar,     prefix     dirs)
@@ -444,6 +475,7 @@ installDirsTemplateEnv dirs =
 
 instance Show PathTemplateVariable where
   show PrefixVar     = "prefix"
+  show PkgKeyVar     = "pkgkey"
   show BindirVar     = "bindir"
   show LibdirVar     = "libdir"
   show LibsubdirVar  = "libsubdir"
@@ -457,6 +489,8 @@ instance Show PathTemplateVariable where
   show CompilerVar   = "compiler"
   show OSVar         = "os"
   show ArchVar       = "arch"
+  show AbiTagVar     = "abitag"
+  show AbiVar        = "abi"
   show ExecutableNameVar = "executablename"
   show TestSuiteNameVar   = "test-suite"
   show TestSuiteResultVar = "result"
@@ -468,6 +502,7 @@ instance Read PathTemplateVariable where
     [ (var, drop (length varStr) s)
     | (varStr, var) <- vars
     , varStr `isPrefixOf` s ]
+    -- NB: order matters! Longer strings first
     where vars = [("prefix",     PrefixVar)
                  ,("bindir",     BindirVar)
                  ,("libdir",     LibdirVar)
@@ -477,11 +512,14 @@ instance Read PathTemplateVariable where
                  ,("docdir",     DocdirVar)
                  ,("htmldir",    HtmldirVar)
                  ,("pkgid",      PkgIdVar)
+                 ,("pkgkey",     PkgKeyVar)
                  ,("pkg",        PkgNameVar)
                  ,("version",    PkgVerVar)
                  ,("compiler",   CompilerVar)
                  ,("os",         OSVar)
                  ,("arch",       ArchVar)
+                 ,("abitag",     AbiTagVar)
+                 ,("abi",        AbiVar)
                  ,("executablename", ExecutableNameVar)
                  ,("test-suite", TestSuiteNameVar)
                  ,("result", TestSuiteResultVar)
@@ -548,7 +586,13 @@ csidl_PROGRAM_FILES = 0x0026
 -- csidl_PROGRAM_FILES_COMMON :: CInt
 -- csidl_PROGRAM_FILES_COMMON = 0x002b
 
-foreign import stdcall unsafe "shlobj.h SHGetFolderPathW"
+#ifdef x86_64_HOST_ARCH
+#define CALLCONV ccall
+#else
+#define CALLCONV stdcall
+#endif
+
+foreign import CALLCONV unsafe "shlobj.h SHGetFolderPathW"
             c_SHGetFolderPath :: Ptr ()
                               -> CInt
                               -> Ptr ()
