@@ -1,3 +1,5 @@
+{-# LANGUAGE DeriveGeneric #-}
+
 -----------------------------------------------------------------------------
 -- |
 -- Module      :  Distribution.Simple.PackageIndex
@@ -12,7 +14,9 @@
 --
 module Distribution.Simple.PackageIndex (
   -- * Package index data type
+  InstalledPackageIndex,
   PackageIndex,
+  FakeMap,
 
   -- * Creating an index
   fromList,
@@ -32,6 +36,7 @@ module Distribution.Simple.PackageIndex (
   -- ** Precise lookups
   lookupInstalledPackageId,
   lookupSourcePackageId,
+  lookupPackageId,
   lookupPackageName,
   lookupDependency,
 
@@ -55,47 +60,92 @@ module Distribution.Simple.PackageIndex (
   dependencyCycles,
   dependencyGraph,
   moduleNameIndex,
+
+  -- ** Variants of special queries supporting fake map
+  fakeLookupInstalledPackageId,
+  brokenPackages',
+  dependencyClosure',
+  reverseDependencyClosure',
+  dependencyInconsistencies',
+  dependencyCycles',
+  dependencyGraph',
   ) where
 
-import Prelude hiding (lookup)
 import Control.Exception (assert)
-import qualified Data.Map as Map
-import Data.Map (Map)
-import qualified Data.Tree  as Tree
-import qualified Data.Graph as Graph
-import qualified Data.Array as Array
 import Data.Array ((!))
+import qualified Data.Array as Array
+import Distribution.Compat.Binary (Binary)
+import qualified Data.Graph as Graph
 import Data.List as List
          ( null, foldl', sort
          , groupBy, sortBy, find, isInfixOf, nubBy, deleteBy, deleteFirstsBy )
 import Data.Monoid (Monoid(..))
+import Data.Map (Map)
+import qualified Data.Map as Map
 import Data.Maybe (isNothing, fromMaybe)
+import qualified Data.Tree  as Tree
+import GHC.Generics (Generic)
+import Prelude hiding (lookup)
 
 import Distribution.Package
          ( PackageName(..), PackageId
          , Package(..), packageName, packageVersion
          , Dependency(Dependency)--, --PackageFixedDeps(..)
-         , InstalledPackageId(..) )
+         , InstalledPackageId(..), PackageInstalled(..) )
 import Distribution.ModuleName
          ( ModuleName )
 import Distribution.InstalledPackageInfo
-         ( InstalledPackageInfo, installedPackageId )
+         ( InstalledPackageInfo )
 import qualified Distribution.InstalledPackageInfo as IPI
 import Distribution.Version
          ( Version, withinRange )
 import Distribution.Simple.Utils (lowercase, comparing, equating)
 
+-- Note [FakeMap]
+-----------------
+-- We'd like to use the PackageIndex defined in this module for
+-- cabal-install's InstallPlan.  However, at the moment, this
+-- data structure is indexed by InstalledPackageId, which we don't
+-- know until after we've compiled a package (whereas InstallPlan
+-- needs to store not-compiled packages in the index.) Eventually,
+-- an InstalledPackageId will be calculatable prior to actually
+-- building the package (making it something of a misnomer), but
+-- at the moment, the "fake installed package ID map" is a workaround
+-- to solve this problem while reusing PackageIndex.  The basic idea
+-- is that, since we don't know what an InstalledPackageId is
+-- beforehand, we just fake up one based on the package ID (it only
+-- needs to be unique for the particular install plan), and fill
+-- it out with the actual generated InstalledPackageId after the
+-- package is successfully compiled.
+--
+-- However, there is a problem: in the index there may be
+-- references using the old package ID, which are now dangling if
+-- we update the InstalledPackageId.  We could map over the entire
+-- index to update these pointers as well (a costly operation), but
+-- instead, we've chosen to parametrize a variety of important functions
+-- by a FakeMap, which records what a fake installed package ID was
+-- actually resolved to post-compilation.  If we do a lookup, we first
+-- check and see if it's a fake ID in the FakeMap.
+--
+-- It's a bit grungy, but we expect this to only be temporary anyway.
+-- (Another possible workaround would have been to *not* update
+-- the installed package ID, but I decided this would be hard to
+-- understand.)
+
+-- | Map from fake installed package IDs to real ones.  See Note [FakeMap]
+type FakeMap = Map InstalledPackageId InstalledPackageId
 
 -- | The collection of information about packages from one or more 'PackageDB's.
+-- These packages generally should have an instance of 'PackageInstalled'
 --
 -- Packages are uniquely identified in by their 'InstalledPackageId', they can
 -- also be efficiently looked up by package name or by name and version.
 --
-data PackageIndex = PackageIndex
+data PackageIndex a = PackageIndex
   -- The primary index. Each InstalledPackageInfo record is uniquely identified
   -- by its InstalledPackageId.
   --
-  !(Map InstalledPackageId InstalledPackageInfo)
+  !(Map InstalledPackageId a)
 
   -- This auxiliary index maps package names (case-sensitively) to all the
   -- versions and instances of that package. This allows us to find all
@@ -108,18 +158,24 @@ data PackageIndex = PackageIndex
   --
   -- FIXME: Clarify what "preference order" means. Check that this invariant is
   -- preserved. See #1463 for discussion.
-  !(Map PackageName (Map Version [InstalledPackageInfo]))
+  !(Map PackageName (Map Version [a]))
 
-  deriving (Show, Read)
+  deriving (Generic, Show, Read)
 
-instance Monoid PackageIndex where
+instance Binary a => Binary (PackageIndex a)
+
+-- | The default package index which contains 'InstalledPackageInfo'.  Normally
+-- use this.
+type InstalledPackageIndex = PackageIndex InstalledPackageInfo
+
+instance PackageInstalled a => Monoid (PackageIndex a) where
   mempty  = PackageIndex Map.empty Map.empty
   mappend = merge
   --save one mappend with empty in the common case:
   mconcat [] = mempty
   mconcat xs = foldr1 mappend xs
 
-invariant :: PackageIndex -> Bool
+invariant :: PackageInstalled a => PackageIndex a -> Bool
 invariant (PackageIndex pids pnames) =
      map installedPackageId (Map.elems pids)
   == sort
@@ -140,9 +196,10 @@ invariant (PackageIndex pids pnames) =
 -- * Internal helpers
 --
 
-mkPackageIndex :: Map InstalledPackageId InstalledPackageInfo
-               -> Map PackageName (Map Version [InstalledPackageInfo])
-               -> PackageIndex
+mkPackageIndex :: PackageInstalled a
+               => Map InstalledPackageId a
+               -> Map PackageName (Map Version [a])
+               -> PackageIndex a
 mkPackageIndex pids pnames = assert (invariant index) index
   where index = PackageIndex pids pnames
 
@@ -156,7 +213,7 @@ mkPackageIndex pids pnames = assert (invariant index) index
 -- If there are duplicates by 'InstalledPackageId' then later ones mask earlier
 -- ones.
 --
-fromList :: [InstalledPackageInfo] -> PackageIndex
+fromList :: PackageInstalled a => [a] -> PackageIndex a
 fromList pkgs = mkPackageIndex pids pnames
   where
     pids      = Map.fromList [ (installedPackageId pkg, pkg) | pkg <- pkgs ]
@@ -188,9 +245,9 @@ fromList pkgs = mkPackageIndex pids pnames
 -- result when we do a lookup by source 'PackageId'. This is the mechanism we
 -- use to prefer user packages over global packages.
 --
-merge :: PackageIndex -> PackageIndex -> PackageIndex
+merge :: PackageInstalled a => PackageIndex a -> PackageIndex a -> PackageIndex a
 merge (PackageIndex pids1 pnames1) (PackageIndex pids2 pnames2) =
-  mkPackageIndex (Map.union pids1 pids2)
+  mkPackageIndex (Map.unionWith (\_ y -> y) pids1 pids2)
                  (Map.unionWith (Map.unionWith mergeBuckets) pnames1 pnames2)
   where
     -- Packages in the second list mask those in the first, however preferred
@@ -204,7 +261,7 @@ merge (PackageIndex pids1 pnames1) (PackageIndex pids2 pnames2) =
 -- This is equivalent to (but slightly quicker than) using 'mappend' or
 -- 'merge' with a singleton index.
 --
-insert :: InstalledPackageInfo -> PackageIndex -> PackageIndex
+insert :: PackageInstalled a => a -> PackageIndex a -> PackageIndex a
 insert pkg (PackageIndex pids pnames) =
     mkPackageIndex pids' pnames'
 
@@ -226,7 +283,7 @@ insert pkg (PackageIndex pids pnames) =
 
 -- | Removes a single installed package from the index.
 --
-deleteInstalledPackageId :: InstalledPackageId -> PackageIndex -> PackageIndex
+deleteInstalledPackageId :: PackageInstalled a => InstalledPackageId -> PackageIndex a -> PackageIndex a
 deleteInstalledPackageId ipkgid original@(PackageIndex pids pnames) =
   case Map.updateLookupWithKey (\_ _ -> Nothing) ipkgid pids of
     (Nothing,     _)     -> original
@@ -248,7 +305,7 @@ deleteInstalledPackageId ipkgid original@(PackageIndex pids pnames) =
 
 -- | Removes all packages with this source 'PackageId' from the index.
 --
-deleteSourcePackageId :: PackageId -> PackageIndex -> PackageIndex
+deleteSourcePackageId :: PackageInstalled a => PackageId -> PackageIndex a -> PackageIndex a
 deleteSourcePackageId pkgid original@(PackageIndex pids pnames) =
   case Map.lookup (packageName pkgid) pnames of
     Nothing     -> original
@@ -268,7 +325,7 @@ deleteSourcePackageId pkgid original@(PackageIndex pids pnames) =
 
 -- | Removes all packages with this (case-sensitive) name from the index.
 --
-deletePackageName :: PackageName -> PackageIndex -> PackageIndex
+deletePackageName :: PackageInstalled a => PackageName -> PackageIndex a -> PackageIndex a
 deletePackageName name original@(PackageIndex pids pnames) =
   case Map.lookup name pnames of
     Nothing     -> original
@@ -291,14 +348,14 @@ deleteDependency (Dependency name verstionRange) =
 
 -- | Get all the packages from the index.
 --
-allPackages :: PackageIndex -> [InstalledPackageInfo]
+allPackages :: PackageIndex a -> [a]
 allPackages (PackageIndex pids _) = Map.elems pids
 
 -- | Get all the packages from the index.
 --
 -- They are grouped by package name (case-sensitively).
 --
-allPackagesByName :: PackageIndex -> [(PackageName, [InstalledPackageInfo])]
+allPackagesByName :: PackageIndex a -> [(PackageName, [a])]
 allPackagesByName (PackageIndex _ pnames) =
   [ (pkgname, concat (Map.elems pvers))
   | (pkgname, pvers) <- Map.toList pnames ]
@@ -307,7 +364,7 @@ allPackagesByName (PackageIndex _ pnames) =
 --
 -- They are grouped by source package id (package name and version).
 --
-allPackagesBySourcePackageId :: PackageIndex -> [(PackageId, [InstalledPackageInfo])]
+allPackagesBySourcePackageId :: PackageInstalled a => PackageIndex a -> [(PackageId, [a])]
 allPackagesBySourcePackageId (PackageIndex _ pnames) =
   [ (packageId ipkg, ipkgs)
   | pvers <- Map.elems pnames
@@ -322,8 +379,8 @@ allPackagesBySourcePackageId (PackageIndex _ pnames) =
 -- Since multiple package DBs mask each other by 'InstalledPackageId',
 -- then we get back at most one package.
 --
-lookupInstalledPackageId :: PackageIndex -> InstalledPackageId
-                         -> Maybe InstalledPackageInfo
+lookupInstalledPackageId :: PackageInstalled a => PackageIndex a -> InstalledPackageId
+                         -> Maybe a
 lookupInstalledPackageId (PackageIndex pids _) pid = Map.lookup pid pids
 
 
@@ -333,7 +390,7 @@ lookupInstalledPackageId (PackageIndex pids _) pid = Map.lookup pid pids
 -- but different 'InstalledPackageId'. They are returned in order of
 -- preference, with the most preferred first.
 --
-lookupSourcePackageId :: PackageIndex -> PackageId -> [InstalledPackageInfo]
+lookupSourcePackageId :: PackageInstalled a => PackageIndex a -> PackageId -> [a]
 lookupSourcePackageId (PackageIndex _ pnames) pkgid =
   case Map.lookup (packageName pkgid) pnames of
     Nothing     -> []
@@ -341,11 +398,18 @@ lookupSourcePackageId (PackageIndex _ pnames) pkgid =
       Nothing   -> []
       Just pkgs -> pkgs -- in preference order
 
+-- | Convenient alias of 'lookupSourcePackageId', but assuming only
+-- one package per package ID.
+lookupPackageId :: PackageInstalled a => PackageIndex a -> PackageId -> Maybe a
+lookupPackageId index pkgid = case lookupSourcePackageId index pkgid  of
+    []    -> Nothing
+    [pkg] -> Just pkg
+    _     -> error "Distribution.Simple.PackageIndex: multiple matches found"
 
 -- | Does a lookup by source package name.
 --
-lookupPackageName :: PackageIndex -> PackageName
-                  -> [(Version, [InstalledPackageInfo])]
+lookupPackageName :: PackageInstalled a => PackageIndex a -> PackageName
+                  -> [(Version, [a])]
 lookupPackageName (PackageIndex _ pnames) name =
   case Map.lookup name pnames of
     Nothing     -> []
@@ -357,8 +421,8 @@ lookupPackageName (PackageIndex _ pnames) name =
 -- We get back any number of versions of the specified package name, all
 -- satisfying the version range constraint.
 --
-lookupDependency :: PackageIndex -> Dependency
-                 -> [(Version, [InstalledPackageInfo])]
+lookupDependency :: PackageInstalled a => PackageIndex a -> Dependency
+                 -> [(Version, [a])]
 lookupDependency (PackageIndex _ pnames) (Dependency name versionRange) =
   case Map.lookup name pnames of
     Nothing    -> []
@@ -382,7 +446,7 @@ lookupDependency (PackageIndex _ pnames) (Dependency name versionRange) =
 -- packages. The list of ambiguous results is split by exact package name. So
 -- it is a non-empty list of non-empty lists.
 --
-searchByName :: PackageIndex -> String -> SearchResult [InstalledPackageInfo]
+searchByName :: PackageInstalled a => PackageIndex a -> String -> SearchResult [a]
 searchByName (PackageIndex _ pnames) name =
   case [ pkgs | pkgs@(PackageName name',_) <- Map.toList pnames
               , lowercase name' == lname ] of
@@ -399,7 +463,7 @@ data SearchResult a = None | Unambiguous a | Ambiguous [a]
 --
 -- That is, all packages that contain the given string in their name.
 --
-searchByNameSubstring :: PackageIndex -> String -> [InstalledPackageInfo]
+searchByNameSubstring :: PackageInstalled a => PackageIndex a -> String -> [a]
 searchByNameSubstring (PackageIndex _ pnames) searchterm =
   [ pkg
   | (PackageName name, pvers) <- Map.toList pnames
@@ -423,11 +487,15 @@ searchByNameSubstring (PackageIndex _ pnames) searchterm =
 -- list of groups of packages where within each group they all depend on each
 -- other, directly or indirectly.
 --
-dependencyCycles :: PackageIndex -> [[InstalledPackageInfo]]
-dependencyCycles index =
+dependencyCycles :: PackageInstalled a => PackageIndex a -> [[a]]
+dependencyCycles = dependencyCycles' Map.empty
+
+-- | Variant of 'dependencyCycles' which accepts a 'FakeMap'.  See Note [FakeMap].
+dependencyCycles' :: PackageInstalled a => FakeMap -> PackageIndex a -> [[a]]
+dependencyCycles' fakeMap index =
   [ vs | Graph.CyclicSCC vs <- Graph.stronglyConnComp adjacencyList ]
   where
-    adjacencyList = [ (pkg, installedPackageId pkg, IPI.depends pkg)
+    adjacencyList = [ (pkg, installedPackageId pkg, fakeInstalledDepends fakeMap pkg)
                     | pkg <- allPackages index ]
 
 
@@ -435,14 +503,21 @@ dependencyCycles index =
 --
 -- Returns such packages along with the dependencies that they're missing.
 --
-brokenPackages :: PackageIndex -> [(InstalledPackageInfo, [InstalledPackageId])]
-brokenPackages index =
+brokenPackages :: PackageInstalled a => PackageIndex a -> [(a, [InstalledPackageId])]
+brokenPackages = brokenPackages' Map.empty
+
+-- | Variant of 'brokenPackages' which accepts a 'FakeMap'.  See Note [FakeMap].
+brokenPackages' :: PackageInstalled a => FakeMap -> PackageIndex a -> [(a, [InstalledPackageId])]
+brokenPackages' fakeMap index =
   [ (pkg, missing)
   | pkg  <- allPackages index
-  , let missing = [ pkg' | pkg' <- IPI.depends pkg
-                         , isNothing (lookupInstalledPackageId index pkg') ]
+  , let missing = [ pkg' | pkg' <- installedDepends pkg
+                         , isNothing (fakeLookupInstalledPackageId fakeMap index pkg') ]
   , not (null missing) ]
 
+-- | Variant of 'lookupInstalledPackageId' which accepts a 'FakeMap'.  See Note [FakeMap].
+fakeLookupInstalledPackageId :: PackageInstalled a => FakeMap -> PackageIndex a -> InstalledPackageId -> Maybe a
+fakeLookupInstalledPackageId fakeMap index pkg = lookupInstalledPackageId index (Map.findWithDefault pkg pkg fakeMap)
 
 -- | Tries to take the transitive closure of the package dependencies.
 --
@@ -452,48 +527,63 @@ brokenPackages index =
 -- * Note that if the result is @Right []@ it is because at least one of
 -- the original given 'PackageId's do not occur in the index.
 --
-dependencyClosure :: PackageIndex
+dependencyClosure :: PackageInstalled a => PackageIndex a
                   -> [InstalledPackageId]
-                  -> Either PackageIndex
-                            [(InstalledPackageInfo, [InstalledPackageId])]
-dependencyClosure index pkgids0 = case closure mempty [] pkgids0 of
+                  -> Either (PackageIndex a)
+                            [(a, [InstalledPackageId])]
+dependencyClosure = dependencyClosure' Map.empty
+
+-- | Variant of 'dependencyClosure' which accepts a 'FakeMap'.  See Note [FakeMap].
+dependencyClosure' :: PackageInstalled a => FakeMap
+                  -> PackageIndex a
+                  -> [InstalledPackageId]
+                  -> Either (PackageIndex a)
+                            [(a, [InstalledPackageId])]
+dependencyClosure' fakeMap index pkgids0 = case closure mempty [] pkgids0 of
   (completed, []) -> Left completed
   (completed, _)  -> Right (brokenPackages completed)
  where
     closure completed failed []             = (completed, failed)
-    closure completed failed (pkgid:pkgids) = case lookupInstalledPackageId index pkgid of
+    closure completed failed (pkgid:pkgids) = case fakeLookupInstalledPackageId fakeMap index pkgid of
       Nothing   -> closure completed (pkgid:failed) pkgids
-      Just pkg  -> case lookupInstalledPackageId completed (installedPackageId pkg) of
+      Just pkg  -> case fakeLookupInstalledPackageId fakeMap completed (installedPackageId pkg) of
         Just _  -> closure completed  failed pkgids
         Nothing -> closure completed' failed pkgids'
           where completed' = insert pkg completed
-                pkgids'    = IPI.depends pkg ++ pkgids
+                pkgids'    = installedDepends pkg ++ pkgids
 
 -- | Takes the transitive closure of the packages reverse dependencies.
 --
 -- * The given 'PackageId's must be in the index.
 --
-reverseDependencyClosure :: PackageIndex
+reverseDependencyClosure :: PackageInstalled a => PackageIndex a
                          -> [InstalledPackageId]
-                         -> [InstalledPackageInfo]
-reverseDependencyClosure index =
+                         -> [a]
+reverseDependencyClosure = reverseDependencyClosure' Map.empty
+
+-- | Variant of 'reverseDependencyClosure' which accepts a 'FakeMap'.  See Note [FakeMap].
+reverseDependencyClosure' :: PackageInstalled a => FakeMap
+                         -> PackageIndex a
+                         -> [InstalledPackageId]
+                         -> [a]
+reverseDependencyClosure' fakeMap index =
     map vertexToPkg
   . concatMap Tree.flatten
   . Graph.dfs reverseDepGraph
   . map (fromMaybe noSuchPkgId . pkgIdToVertex)
 
   where
-    (depGraph, vertexToPkg, pkgIdToVertex) = dependencyGraph index
+    (depGraph, vertexToPkg, pkgIdToVertex) = dependencyGraph' fakeMap index
     reverseDepGraph = Graph.transposeG depGraph
     noSuchPkgId = error "reverseDependencyClosure: package is not in the graph"
 
-topologicalOrder :: PackageIndex -> [InstalledPackageInfo]
+topologicalOrder :: PackageInstalled a => PackageIndex a -> [a]
 topologicalOrder index = map toPkgId
                        . Graph.topSort
                        $ graph
   where (graph, toPkgId, _) = dependencyGraph index
 
-reverseTopologicalOrder :: PackageIndex -> [InstalledPackageInfo]
+reverseTopologicalOrder :: PackageInstalled a => PackageIndex a -> [a]
 reverseTopologicalOrder index = map toPkgId
                               . Graph.topSort
                               . Graph.transposeG
@@ -505,20 +595,28 @@ reverseTopologicalOrder index = map toPkgId
 -- Dependencies on other packages that are not in the index are discarded.
 -- You can check if there are any such dependencies with 'brokenPackages'.
 --
-dependencyGraph :: PackageIndex
+dependencyGraph :: PackageInstalled a => PackageIndex a
                 -> (Graph.Graph,
-                    Graph.Vertex -> InstalledPackageInfo,
+                    Graph.Vertex -> a,
                     InstalledPackageId -> Maybe Graph.Vertex)
-dependencyGraph index = (graph, vertex_to_pkg, id_to_vertex)
+dependencyGraph = dependencyGraph' Map.empty
+
+-- | Variant of 'dependencyGraph' which accepts a 'FakeMap'.  See Note [FakeMap].
+dependencyGraph' :: PackageInstalled a => FakeMap
+                -> PackageIndex a
+                -> (Graph.Graph,
+                    Graph.Vertex -> a,
+                    InstalledPackageId -> Maybe Graph.Vertex)
+dependencyGraph' fakeMap index = (graph, vertex_to_pkg, id_to_vertex)
   where
     graph = Array.listArray bounds
-              [ [ v | Just v <- map id_to_vertex (IPI.depends pkg) ]
+              [ [ v | Just v <- map id_to_vertex (installedDepends pkg) ]
               | pkg <- pkgs ]
 
     pkgs             = sortBy (comparing packageId) (allPackages index)
     vertices         = zip (map installedPackageId pkgs) [0..]
     vertex_map       = Map.fromList vertices
-    id_to_vertex pid = Map.lookup pid vertex_map
+    id_to_vertex pid = Map.lookup (Map.findWithDefault pid pid fakeMap) vertex_map
 
     vertex_to_pkg vertex = pkgTable ! vertex
 
@@ -536,9 +634,14 @@ dependencyGraph index = (graph, vertex_to_pkg, id_to_vertex)
 -- depend on it and the versions they require. These are guaranteed to be
 -- distinct.
 --
-dependencyInconsistencies :: PackageIndex
+dependencyInconsistencies :: PackageInstalled a => PackageIndex a
                           -> [(PackageName, [(PackageId, Version)])]
-dependencyInconsistencies index =
+dependencyInconsistencies = dependencyInconsistencies' Map.empty
+
+-- | Variant of 'dependencyInconsistencies' which accepts a 'FakeMap'.  See Note [FakeMap].
+dependencyInconsistencies' :: PackageInstalled a => FakeMap -> PackageIndex a
+                          -> [(PackageName, [(PackageId, Version)])]
+dependencyInconsistencies' fakeMap index =
   [ (name, [ (pid,packageVersion dep) | (dep,pids) <- uses, pid <- pids])
   | (name, ipid_map) <- Map.toList inverseIndex
   , let uses = Map.elems ipid_map
@@ -548,29 +651,42 @@ dependencyInconsistencies index =
         --   for each package with that name,
         --     the InstalledPackageInfo and the package Ids of packages
         --     that depend on it.
-        inverseIndex :: Map PackageName
-                            (Map InstalledPackageId
-                                 (InstalledPackageInfo, [PackageId]))
         inverseIndex = Map.fromListWith (Map.unionWith (\(a,b) (_,b') -> (a,b++b')))
           [ (packageName dep,
              Map.fromList [(ipid,(dep,[packageId pkg]))])
           | pkg <- allPackages index
-          , ipid <- IPI.depends pkg
-          , Just dep <- [lookupInstalledPackageId index ipid]
+          , ipid <- fakeInstalledDepends fakeMap pkg
+          , Just dep <- [fakeLookupInstalledPackageId fakeMap index ipid]
           ]
 
-        reallyIsInconsistent :: [InstalledPackageInfo] -> Bool
+        reallyIsInconsistent :: PackageInstalled a => [a] -> Bool
         reallyIsInconsistent []       = False
         reallyIsInconsistent [_p]     = False
         reallyIsInconsistent [p1, p2] =
-             installedPackageId p1 `notElem` IPI.depends p2
-          && installedPackageId p2 `notElem` IPI.depends p1
+          let pid1 = installedPackageId p1
+              pid2 = installedPackageId p2
+          in Map.findWithDefault pid1 pid1 fakeMap `notElem` fakeInstalledDepends fakeMap p2
+          && Map.findWithDefault pid2 pid2 fakeMap `notElem` fakeInstalledDepends fakeMap p1
         reallyIsInconsistent _ = True
 
+-- | Variant of 'installedDepends' which accepts a 'FakeMap'.  See Note [FakeMap].
+fakeInstalledDepends :: PackageInstalled a => FakeMap -> a -> [InstalledPackageId]
+fakeInstalledDepends fakeMap = map (\pid -> Map.findWithDefault pid pid fakeMap) . installedDepends
 
-moduleNameIndex :: PackageIndex -> Map ModuleName [InstalledPackageInfo]
+-- | A rough approximation of GHC's module finder, takes a 'InstalledPackageIndex' and
+-- turns it into a map from module names to their source packages.  It's used to
+-- initialize the @build-deps@ field in @cabal init@.
+moduleNameIndex :: InstalledPackageIndex -> Map ModuleName [InstalledPackageInfo]
 moduleNameIndex index =
-  Map.fromListWith (++)
-    [ (moduleName, [pkg])
-    | pkg        <- allPackages index
-    , moduleName <- IPI.exposedModules pkg ]
+  Map.fromListWith (++) $ do
+    pkg <- allPackages index
+    IPI.ExposedModule m reexport _ <- IPI.exposedModules pkg
+    case reexport of
+        Nothing -> return (m, [pkg])
+        Just (IPI.OriginalModule _ m') | m == m'   -> []
+                                       | otherwise -> return (m', [pkg])
+        -- The heuristic is this: we want to prefer the original package
+        -- which originally exported a module.  However, if a reexport
+        -- also *renamed* the module (m /= m'), then we have to use the
+        -- downstream package, since the upstream package has the wrong
+        -- module name!
