@@ -11,20 +11,23 @@ module Distribution.Simple.Program.GHC (
 
   ) where
 
+import Distribution.Simple.GHC.ImplInfo ( getImplInfo, GhcImplInfo(..) )
 import Distribution.Package
+import Distribution.PackageDescription hiding (Flag)
 import Distribution.ModuleName
 import Distribution.Simple.Compiler hiding (Flag)
 import Distribution.Simple.Setup    ( Flag(..), flagToMaybe, fromFlagOrDefault,
                                       flagToList )
---import Distribution.Simple.LocalBuildInfo
 import Distribution.Simple.Program.Types
 import Distribution.Simple.Program.Run
 import Distribution.Text
 import Distribution.Verbosity
-import Distribution.Version
+import Distribution.Utils.NubList   ( NubListR, fromNubListR )
 import Language.Haskell.Extension   ( Language(..), Extension(..) )
 
+import qualified Data.Map as M
 import Data.Monoid
+import Data.List ( intercalate )
 
 -- | A structured set of GHC options/flags
 --
@@ -35,20 +38,20 @@ data GhcOptions = GhcOptions {
 
   -- | Any extra options to pass directly to ghc. These go at the end and hence
   -- override other stuff.
-  ghcOptExtra         :: [String],
+  ghcOptExtra         :: NubListR String,
 
   -- | Extra default flags to pass directly to ghc. These go at the beginning
   -- and so can be overridden by other stuff.
-  ghcOptExtraDefault  :: [String],
+  ghcOptExtraDefault  :: NubListR String,
 
   -----------------------
   -- Inputs and outputs
 
   -- | The main input files; could be .hs, .hi, .c, .o, depending on mode.
-  ghcOptInputFiles    :: [FilePath],
+  ghcOptInputFiles    :: NubListR FilePath,
 
   -- | The names of input Haskell modules, mainly for @--make@ mode.
-  ghcOptInputModules  :: [ModuleName],
+  ghcOptInputModules  :: NubListR ModuleName,
 
   -- | Location for output file; the @ghc -o@ flag.
   ghcOptOutputFile    :: Flag FilePath,
@@ -62,66 +65,73 @@ data GhcOptions = GhcOptions {
   ghcOptSourcePathClear :: Flag Bool,
 
   -- | Search path for Haskell source files; the @ghc -i@ flag.
-  ghcOptSourcePath    :: [FilePath],
+  ghcOptSourcePath    :: NubListR FilePath,
 
   -------------
   -- Packages
 
-  -- | The package name the modules will belong to; the @ghc -package-name@ flag
-  ghcOptPackageName   :: Flag PackageId,
+  -- | The package key the modules will belong to; the @ghc -this-package-key@
+  -- flag.
+  ghcOptPackageKey   :: Flag PackageKey,
 
-  -- | GHC package databases to use, the @ghc -package-conf@ flag
+  -- | GHC package databases to use, the @ghc -package-conf@ flag.
   ghcOptPackageDBs    :: PackageDBStack,
 
   -- | The GHC packages to use. For compatability with old and new ghc, this
   -- requires both the short and long form of the package id;
   -- the @ghc -package@ or @ghc -package-id@ flags.
-  ghcOptPackages      :: [(InstalledPackageId, PackageId)],
+  ghcOptPackages      ::
+    NubListR (InstalledPackageId, PackageId, ModuleRenaming),
 
   -- | Start with a clean package set; the @ghc -hide-all-packages@ flag
   ghcOptHideAllPackages :: Flag Bool,
 
-  -- | Don't automatically link in Haskell98 etc; the @ghc -no-auto-link-packages@ flag.
+  -- | Don't automatically link in Haskell98 etc; the @ghc
+  -- -no-auto-link-packages@ flag.
   ghcOptNoAutoLinkPackages :: Flag Bool,
 
+  -- | What packages are implementing the signatures
+  ghcOptSigOf :: [(ModuleName, (PackageKey, ModuleName))],
+
   -----------------
   -- Linker stuff
 
   -- | Names of libraries to link in; the @ghc -l@ flag.
-  ghcOptLinkLibs      :: [FilePath],
+  ghcOptLinkLibs      :: NubListR FilePath,
 
   -- | Search path for libraries to link in; the @ghc -L@ flag.
-  ghcOptLinkLibPath  :: [FilePath],
+  ghcOptLinkLibPath  :: NubListR FilePath,
 
   -- | Options to pass through to the linker; the @ghc -optl@ flag.
-  ghcOptLinkOptions   :: [String],
+  ghcOptLinkOptions   :: NubListR String,
 
   -- | OSX only: frameworks to link in; the @ghc -framework@ flag.
-  ghcOptLinkFrameworks :: [String],
+  ghcOptLinkFrameworks :: NubListR String,
 
   -- | Don't do the link step, useful in make mode; the @ghc -no-link@ flag.
   ghcOptNoLink :: Flag Bool,
 
-  -- | Don't link in the normal RTS @main@ entry point; the @ghc -no-hs-main@ flag.
+  -- | Don't link in the normal RTS @main@ entry point; the @ghc -no-hs-main@
+  -- flag.
   ghcOptLinkNoHsMain :: Flag Bool,
 
   --------------------
   -- C and CPP stuff
 
   -- | Options to pass through to the C compiler; the @ghc -optc@ flag.
-  ghcOptCcOptions     :: [String],
+  ghcOptCcOptions     :: NubListR String,
 
   -- | Options to pass through to CPP; the @ghc -optP@ flag.
-  ghcOptCppOptions    :: [String],
+  ghcOptCppOptions    :: NubListR String,
 
   -- | Search path for CPP includes like header files; the @ghc -I@ flag.
-  ghcOptCppIncludePath :: [FilePath],
+  ghcOptCppIncludePath :: NubListR FilePath,
 
   -- | Extra header files to include at CPP stage; the @ghc -optP-include@ flag.
-  ghcOptCppIncludes    :: [FilePath],
+  ghcOptCppIncludes    :: NubListR FilePath,
 
   -- | Extra header files to include for old-style FFI; the @ghc -#include@ flag.
-  ghcOptFfiIncludes    :: [FilePath],
+  ghcOptFfiIncludes    :: NubListR FilePath,
 
   ----------------------------
   -- Language and extensions
@@ -130,11 +140,11 @@ data GhcOptions = GhcOptions {
   ghcOptLanguage      :: Flag Language,
 
   -- | The language extensions; the @ghc -X@ flag.
-  ghcOptExtensions    :: [Extension],
+  ghcOptExtensions    :: NubListR Extension,
 
   -- | A GHC version-dependent mapping of extensions to flags. This must be
   -- set to be able to make use of the 'ghcOptExtensions'.
-  ghcOptExtensionMap    :: [(Extension, String)],
+  ghcOptExtensionMap    :: M.Map Extension String,
 
   ----------------
   -- Compilation
@@ -142,6 +152,9 @@ data GhcOptions = GhcOptions {
   -- | What optimisation level to use; the @ghc -O@ flag.
   ghcOptOptimisation  :: Flag GhcOptimisation,
 
+    -- | Emit debug info; the @ghc -g@ flag.
+  ghcOptDebugInfo  :: Flag Bool,
+
   -- | Compile in profiling mode; the @ghc -prof@ flag.
   ghcOptProfilingMode :: Flag Bool,
 
@@ -149,13 +162,16 @@ data GhcOptions = GhcOptions {
   ghcOptSplitObjs     :: Flag Bool,
 
   -- | Run N jobs simultaneously (if possible).
-  ghcOptNumJobs       :: Flag Int,
+  ghcOptNumJobs       :: Flag (Maybe Int),
+
+  -- | Enable coverage analysis; the @ghc -fhpc -hpcdir@ flags.
+  ghcOptHPCDir        :: Flag FilePath,
 
   ----------------
   -- GHCi
 
   -- | Extra GHCi startup scripts; the @-ghci-script@ flag
-  ghcOptGHCiScripts    :: [FilePath],
+  ghcOptGHCiScripts    :: NubListR FilePath,
 
   ------------------------
   -- Redirecting outputs
@@ -176,6 +192,7 @@ data GhcOptions = GhcOptions {
   ghcOptShared        :: Flag Bool,
   ghcOptFPic          :: Flag Bool,
   ghcOptDylibName     :: Flag String,
+  ghcOptRPaths        :: NubListR FilePath,
 
   ---------------
   -- Misc flags
@@ -220,12 +237,11 @@ ghcInvocation :: ConfiguredProgram -> Compiler -> GhcOptions -> ProgramInvocatio
 ghcInvocation prog comp opts =
     programInvocation prog (renderGhcOptions comp opts)
 
-
 renderGhcOptions :: Compiler -> GhcOptions -> [String]
 renderGhcOptions comp opts
-  | compilerFlavor comp /= GHC =
+  | compilerFlavor comp `notElem` [GHC, GHCJS] =
     error $ "Distribution.Simple.Program.GHC.renderGhcOptions: "
-    ++ "compiler flavor must be 'GHC'!"
+    ++ "compiler flavor must be 'GHC' or 'GHCJS'!"
   | otherwise =
   concat
   [ case flagToMaybe (ghcOptMode opts) of
@@ -247,7 +263,8 @@ renderGhcOptions comp opts
 
   , maybe [] verbosityOpts (flagToMaybe (ghcOptVerbosity opts))
 
-  , [ "-fbuilding-cabal-package" | flagBool ghcOptCabal, ver >= [6,11] ]
+  , [ "-fbuilding-cabal-package" | flagBool ghcOptCabal
+                                 , flagBuildingCabalPkg implInfo ]
 
   ----------------
   -- Compilation
@@ -259,15 +276,20 @@ renderGhcOptions comp opts
       Just GhcMaximumOptimisation     -> ["-O2"]
       Just (GhcSpecialOptimisation s) -> ["-O" ++ s] -- eg -Odph
 
+  , [ "-g" | flagDebugInfo implInfo && flagBool ghcOptDebugInfo ]
+
   , [ "-prof" | flagBool ghcOptProfilingMode ]
 
   , [ "-split-objs" | flagBool ghcOptSplitObjs ]
 
+  , case flagToMaybe (ghcOptHPCDir opts) of
+      Nothing -> []
+      Just hpcdir -> ["-fhpc", "-hpcdir", hpcdir]
 
   , if parmakeSupported comp
-    then
-      let numJobs = fromFlagOrDefault 1 (ghcOptNumJobs opts)
-      in if numJobs > 1 then ["-j" ++ show numJobs] else []
+    then case ghcOptNumJobs opts of
+      NoFlag  -> []
+      Flag n  -> ["-j" ++ maybe "" show n]
     else []
 
   --------------------
@@ -290,10 +312,12 @@ renderGhcOptions comp opts
   , concat [ ["-hisuf",   suf] | suf <- flag ghcOptHiSuffix  ]
   , concat [ ["-dynosuf", suf] | suf <- flag ghcOptDynObjSuffix ]
   , concat [ ["-dynhisuf",suf] | suf <- flag ghcOptDynHiSuffix  ]
-  , concat [ ["-outputdir", dir] | dir <- flag ghcOptOutputDir, ver >= [6,10] ]
+  , concat [ ["-outputdir", dir] | dir <- flag ghcOptOutputDir
+                                 , flagOutputDir implInfo ]
   , concat [ ["-odir",    dir] | dir <- flag ghcOptObjDir ]
   , concat [ ["-hidir",   dir] | dir <- flag ghcOptHiDir  ]
-  , concat [ ["-stubdir", dir] | dir <- flag ghcOptStubDir, ver >= [6,8] ]
+  , concat [ ["-stubdir", dir] | dir <- flag ghcOptStubDir
+                               , flagStubdir implInfo ]
 
   -----------------------
   -- Source search path
@@ -306,8 +330,10 @@ renderGhcOptions comp opts
 
   , [ "-I"    ++ dir | dir <- flags ghcOptCppIncludePath ]
   , [ "-optP" ++ opt | opt <- flags ghcOptCppOptions ]
-  , concat [ [ "-optP-include", "-optP" ++ inc] | inc <- flags ghcOptCppIncludes ]
-  , [ "-#include \"" ++ inc ++ "\"" | inc <- flags ghcOptFfiIncludes, ver < [6,11] ]
+  , concat [ [ "-optP-include", "-optP" ++ inc]
+           | inc <- flags ghcOptCppIncludes ]
+  , [ "-#include \"" ++ inc ++ "\""
+    | inc <- flags ghcOptFfiIncludes, flagFfiIncludes implInfo ]
   , [ "-optc" ++ opt | opt <- flags ghcOptCcOptions ]
 
   -----------------
@@ -318,45 +344,64 @@ renderGhcOptions comp opts
   , ["-L" ++ dir     | dir <- flags ghcOptLinkLibPath ]
   , concat [ ["-framework", fmwk] | fmwk <- flags ghcOptLinkFrameworks ]
   , [ "-no-hs-main"  | flagBool ghcOptLinkNoHsMain ]
+  , [ "-dynload deploy" | not (null (flags ghcOptRPaths)) ]
+  , concat [ [ "-optl-Wl,-rpath," ++ dir]
+           | dir <- flags ghcOptRPaths ]
 
   -------------
   -- Packages
 
-  , concat [ ["-package-name", display pkgid] | pkgid <- flag ghcOptPackageName ]
+  , concat [ [if packageKeySupported comp
+                then "-this-package-key"
+                else "-package-name", display pkgid]
+             | pkgid <- flag ghcOptPackageKey ]
 
   , [ "-hide-all-packages"     | flagBool ghcOptHideAllPackages ]
   , [ "-no-auto-link-packages" | flagBool ghcOptNoAutoLinkPackages ]
 
-  , packageDbArgs version (flags ghcOptPackageDBs)
+  , packageDbArgs implInfo (ghcOptPackageDBs opts)
 
-  , concat $ if ver >= [6,11]
-      then [ ["-package-id", display ipkgid] | (ipkgid,_) <- flags ghcOptPackages ]
-      else [ ["-package",    display  pkgid] | (_,pkgid)  <- flags ghcOptPackages ]
+  , if null (ghcOptSigOf opts)
+        then []
+        else "-sig-of"
+             : intercalate "," (map (\(n,(p,m)) -> display n ++ " is "
+                                                ++ display p ++ ":"
+                                                ++ display m)
+                                    (ghcOptSigOf opts))
+             : []
+
+  , concat $ if flagPackageId implInfo
+      then let space "" = ""
+               space xs = ' ' : xs
+           in [ ["-package-id", display ipkgid ++ space (display rns)]
+              | (ipkgid,_,rns) <- flags ghcOptPackages ]
+      else [ ["-package",    display  pkgid]
+           | (_,pkgid,_)  <- flags ghcOptPackages ]
 
   ----------------------------
   -- Language and extensions
 
-  , if ver >= [7]
+  , if supportsHaskell2010 implInfo
     then [ "-X" ++ display lang | lang <- flag ghcOptLanguage ]
     else []
 
-  , [ case lookup ext (ghcOptExtensionMap opts) of
+  , [ case M.lookup ext (ghcOptExtensionMap opts) of
         Just arg -> arg
         Nothing  -> error $ "Distribution.Simple.Program.GHC.renderGhcOptions: "
                           ++ display ext ++ " not present in ghcOptExtensionMap."
-    | ext <- ghcOptExtensions opts ]
+    | ext <- flags ghcOptExtensions ]
 
   ----------------
   -- GHCi
 
   , concat [ [ "-ghci-script", script ] | script <- flags  ghcOptGHCiScripts
-                                        , ver >= [7,2] ]
+                                        , flagGhciScript implInfo ]
 
   ---------------
   -- Inputs
 
   , [ display modu | modu <- flags ghcOptInputModules ]
-  , ghcOptInputFiles opts
+  , flags ghcOptInputFiles
 
   , concat [ [ "-o",    out] | out <- flag ghcOptOutputFile ]
   , concat [ [ "-dyno", out] | out <- flag ghcOptOutputDynFile ]
@@ -364,18 +409,17 @@ renderGhcOptions comp opts
   ---------------
   -- Extra
 
-  , ghcOptExtra opts
+  , flags ghcOptExtra
 
   ]
 
 
   where
+    implInfo     = getImplInfo comp
     flag     flg = flagToList (flg opts)
-    flags    flg = flg opts
+    flags    flg = fromNubListR . flg $ opts
     flagBool flg = fromFlagOrDefault False (flg opts)
 
-    version@(Version ver _) = compilerVersion comp
-
 verbosityOpts :: Verbosity -> [String]
 verbosityOpts verbosity
   | verbosity >= deafening = ["-v"]
@@ -383,8 +427,8 @@ verbosityOpts verbosity
   | otherwise              = ["-w", "-v0"]
 
 
-packageDbArgs :: Version -> PackageDBStack -> [String]
-packageDbArgs (Version ver _) dbstack = case dbstack of
+packageDbArgs :: GhcImplInfo -> PackageDBStack -> [String]
+packageDbArgs implInfo dbstack = case dbstack of
   (GlobalPackageDB:UserPackageDB:dbs) -> concatMap specific dbs
   (GlobalPackageDB:dbs)               -> ("-no-user-" ++ packageDbFlag)
                                        : concatMap specific dbs
@@ -394,9 +438,8 @@ packageDbArgs (Version ver _) dbstack = case dbstack of
     specific _ = ierror
     ierror     = error $ "internal error: unexpected package db stack: "
                       ++ show dbstack
-
     packageDbFlag
-      | ver < [7,5]
+      | flagPackageConf implInfo
       = "package-conf"
       | otherwise
       = "package-db"
@@ -416,11 +459,12 @@ instance Monoid GhcOptions where
     ghcOptOutputDynFile      = mempty,
     ghcOptSourcePathClear    = mempty,
     ghcOptSourcePath         = mempty,
-    ghcOptPackageName        = mempty,
+    ghcOptPackageKey         = mempty,
     ghcOptPackageDBs         = mempty,
     ghcOptPackages           = mempty,
     ghcOptHideAllPackages    = mempty,
     ghcOptNoAutoLinkPackages = mempty,
+    ghcOptSigOf              = mempty,
     ghcOptLinkLibs           = mempty,
     ghcOptLinkLibPath        = mempty,
     ghcOptLinkOptions        = mempty,
@@ -436,9 +480,11 @@ instance Monoid GhcOptions where
     ghcOptExtensions         = mempty,
     ghcOptExtensionMap       = mempty,
     ghcOptOptimisation       = mempty,
+    ghcOptDebugInfo          = mempty,
     ghcOptProfilingMode      = mempty,
     ghcOptSplitObjs          = mempty,
     ghcOptNumJobs            = mempty,
+    ghcOptHPCDir             = mempty,
     ghcOptGHCiScripts        = mempty,
     ghcOptHiSuffix           = mempty,
     ghcOptObjSuffix          = mempty,
@@ -452,6 +498,7 @@ instance Monoid GhcOptions where
     ghcOptShared             = mempty,
     ghcOptFPic               = mempty,
     ghcOptDylibName          = mempty,
+    ghcOptRPaths             = mempty,
     ghcOptVerbosity          = mempty,
     ghcOptCabal              = mempty
   }
@@ -465,11 +512,12 @@ instance Monoid GhcOptions where
     ghcOptOutputDynFile      = combine ghcOptOutputDynFile,
     ghcOptSourcePathClear    = combine ghcOptSourcePathClear,
     ghcOptSourcePath         = combine ghcOptSourcePath,
-    ghcOptPackageName        = combine ghcOptPackageName,
+    ghcOptPackageKey         = combine ghcOptPackageKey,
     ghcOptPackageDBs         = combine ghcOptPackageDBs,
     ghcOptPackages           = combine ghcOptPackages,
     ghcOptHideAllPackages    = combine ghcOptHideAllPackages,
     ghcOptNoAutoLinkPackages = combine ghcOptNoAutoLinkPackages,
+    ghcOptSigOf              = combine ghcOptSigOf,
     ghcOptLinkLibs           = combine ghcOptLinkLibs,
     ghcOptLinkLibPath        = combine ghcOptLinkLibPath,
     ghcOptLinkOptions        = combine ghcOptLinkOptions,
@@ -485,9 +533,11 @@ instance Monoid GhcOptions where
     ghcOptExtensions         = combine ghcOptExtensions,
     ghcOptExtensionMap       = combine ghcOptExtensionMap,
     ghcOptOptimisation       = combine ghcOptOptimisation,
+    ghcOptDebugInfo          = combine ghcOptDebugInfo,
     ghcOptProfilingMode      = combine ghcOptProfilingMode,
     ghcOptSplitObjs          = combine ghcOptSplitObjs,
     ghcOptNumJobs            = combine ghcOptNumJobs,
+    ghcOptHPCDir             = combine ghcOptHPCDir,
     ghcOptGHCiScripts        = combine ghcOptGHCiScripts,
     ghcOptHiSuffix           = combine ghcOptHiSuffix,
     ghcOptObjSuffix          = combine ghcOptObjSuffix,
@@ -501,6 +551,7 @@ instance Monoid GhcOptions where
     ghcOptShared             = combine ghcOptShared,
     ghcOptFPic               = combine ghcOptFPic,
     ghcOptDylibName          = combine ghcOptDylibName,
+    ghcOptRPaths             = combine ghcOptRPaths,
     ghcOptVerbosity          = combine ghcOptVerbosity,
     ghcOptCabal              = combine ghcOptCabal
   }
